#include "RenderTraverser.h"
#include "glInterface.h"
#include "Timers.h"
#include "Camera.h"
#include "SceneEntity.h"
#include "RenderState.h"


using namespace std;


namespace CHCDemo
{

	
void TraversalStatistics::Reset()
{
	// reset statistics
	mNumTraversedNodes = 0;
	mNumQueryCulledNodes = 0;
	mNumFrustumCulledNodes = 0;
	mNumRenderedGeometry = 0;
	mNumRenderedTriangles = 0;
	mNumPreviouslyVisibleNodeQueries = 0;
	mNumIssuedQueries = 0;
	mRenderTime = 0;
	mNumStateChanges = 0;
}



/******************************************************/
/*           RenderTraverser implementation           */
/******************************************************/


RenderTraverser::RenderTraverser():  
mVisibilityThreshold(0),
mBvh(NULL), 
mUseOptimization(false),
mFrameId(-1),
mUseRenderQueue(false),
mAssumedVisibleFrames(10),
mMaxBatchSize(50)
{
}


RenderTraverser::~RenderTraverser()
{
	mQueryHandler.DestroyQueries();
}


void RenderTraverser::EnqueueNode(BvhNode *node)
{
	mBvh->CalcDistance(node);
	mDistanceQueue.push(node);
}


void RenderTraverser::TraverseNode(BvhNode *node)
{
	++ mStats.mNumTraversedNodes;

	if (node->IsVirtualLeaf())
	{
		RenderNode(node);
	}
	else 
	{
		// for non leafs this renders only the bounding volume (if the flag is set)
		BvhInterior *interior = static_cast<BvhInterior *>(node);

		EnqueueNode(interior->GetFront());
		EnqueueNode(interior->GetBack());
	}
}


void RenderTraverser::RenderNode(BvhNode *node)
{
	if (node->GetLastRenderedFrame() != mFrameId)
	{
		//cout << "big error!!" << endl;
		node->SetLastRenderedFrame(mFrameId);

		if (mUseRenderQueue)
		{
			int geometrySize;
			SceneEntity **geom = mBvh->GetGeometry(node, geometrySize);
			mRenderQueue.Enqueue(geom, geometrySize);
			mStats.mNumRenderedGeometry += geometrySize;
		}
		else
		{
			if (mRenderState->SetState(RenderState::RENDER))
				++ mStats.mNumStateChanges;

			mStats.mNumRenderedGeometry += mBvh->Render(node, mRenderState);
		}		
	}
}


void RenderTraverser::SetHierarchy(Bvh *bvh)
{
	mBvh = bvh;
}


void RenderTraverser::SetRenderState(RenderState *state)
{
	mRenderState = state;
	mRenderQueue.SetRenderState(state);
}


OcclusionQuery *RenderTraverser::IssueOcclusionQuery(BvhNode *node, bool wasVisible)
{
	++ mStats.mNumIssuedQueries;

	if (mUseRenderQueue && (mRenderState->GetMode() == RenderState::RENDER))
	{
		mRenderQueue.Render();
		mRenderQueue.Clear();
	}

	OcclusionQuery *query = mQueryHandler.RequestQuery();

	query->SetNode(node);

	query->BeginQuery();

	if (wasVisible)
		++ mStats.mNumPreviouslyVisibleNodeQueries;

	// if this node is a previous visible leaf:
	// leaves will be rendered anyway => we can also test with the real geometry 
	if (wasVisible && mUseOptimization)
	{
		RenderNode(node);
	}
	else
	{
		// change to query mode and render box
		if (mRenderState->SetState(RenderState::QUERY))
			++ mStats.mNumStateChanges;

		mBvh->RenderBoundingBox(node);
	}

	query->EndQuery();

	return query;
}


void RenderTraverser::RenderScene()
{
	InitTiming();
	long t1, t2;

	t1 = GetTime();

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	++ mFrameId;

	mBvh->InitFrame();

	mStats.Reset();
	mQueryHandler.ResetQueries();
	
	EnqueueNode(mBvh->GetRoot());


	///////////
	//-- the actual rendering algorithm

	Traverse();

	mRenderState->Reset();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);

	t2 = GetTime();
	mStats.mRenderTime = TimeDiff(t1, t2);
}


void RenderTraverser::SetUseRenderQueue(bool useRenderQueue)
{
	mUseRenderQueue = useRenderQueue;
	std::cout << "using render queue: " << mUseRenderQueue << std::endl;
}




void RenderTraverser::SetVisibilityThreshold(int threshold)
{
	mVisibilityThreshold = threshold;
}


void RenderTraverser::SetUseOptimization(bool useOptimization)
{
	mUseOptimization = useOptimization;
}


}
